#include "Loader.hpp"

#include "Encoding.hpp"
#include "Error.hpp"

#include <string.h>
#include <stdio.h>

namespace SoftWire
{
	Loader::Link::Link()
	{
		next = 0;
	}

	Loader::Link::~Link()
	{
		delete next;
	}

	void Loader::Link::attachNew(const Encoding &encoding)
	{
		if(next)
		{
			next->attachNew(encoding);
		}
		else
		{
			this->encoding = encoding;

			next = new Link();
		}
	}

	Loader::Loader()
	{
		objectCode = 0;
		entry = 0;
	}

	Loader::~Loader()
	{
		delete[] objectCode;
		delete entry;
	}

	const void *Loader::callable()
	{
		if(!objectCode)
		{
			loadCode();
		}

		if(!objectCode)
		{
			throw Error::INTERNAL;
		}

		return objectCode;
	}

	void Loader::addEncoding(const Encoding &encoding)
	{
		if(!entry)
		{
			entry = new Link();
		}

		entry->attachNew(encoding);
	}

	void Loader::loadCode()
	{
		int length = codeLength();

		objectCode = new byte[length];
		byte *code = objectCode;

		Link *link = entry;

		while(link)
		{
			if(link->encoding.getReference())
			{
				int offset = resolveLabel(link->encoding.getReference()) - code - link->encoding.length();
				link->encoding.setJumpOffset(offset);
			}

			code += link->encoding.writeCode(code);

			link = link->next;
		}

		#ifndef NDEBUG
				printf("\n");
		#endif
	}

	byte *Loader::resolveLabel(const char *label) const
	{
		int offset = 0;
		const Link *link = entry;

		while(link)
		{
			if(link->encoding.getLabel() && strcmp(link->encoding.getLabel(), label) == 0)
			{
				return objectCode + offset;
			}
			else
			{
				offset += link->encoding.length();
			}

			link = link->next;
		}

		throw Error("Undefined label '%s'", label);

		return 0;
	}

	int Loader::codeLength() const
	{
		const Link *link = entry;
		int length = 0;

		while(link)
		{
			length += link->encoding.length();

			link = link->next;
		}
		
		return length;
	}
}